import pickle
import re
import warnings
from os import listdir, system
from os.path import isfile, join

# Used to replace wrkallocs (Probably can be removed when trying to do thigs for other models than NEMO)
from ReplaceWrkAllocFunction import replace_wrkalloc


# Functions
def split_semicolon(string):
    # Trying to use a rarely used character ({) for the temporal replacement
    if string.count(";"):
        # In the case where there's quotes in the text, we should ignore the exclamations inside
        if not string.count("'"):
            return string.split(";")
        else:
            # Looking for
            scan_quotation = [string[:i].count("'") for i, c in enumerate(string)]
            splitted_text = [c for c in string]
            splitted_text = ["{" if c == ";" and scan_quotation[i] % 2 == 1 else c for i, c in
                             enumerate(splitted_text)]
            string = "".join(splitted_text)
            return string.split(";")

    raise AssertionError("string must contain semi-colon")


def strip_call_arguments(_line):
    # Description: Given a string with a comma separated arguments,
    # extract the list of arguments getting rid of the parenthesis
    # Input: String with subroutine declaration
    # Output: List with arguments
    op = [i for i, x in enumerate(_line) if x == "("]
    cp = [i for i, x in enumerate(_line) if x == ")"]
    _line = _line[op[0] + 1:cp[-1]]
    op = [i for i, x in enumerate(_line) if x == "("]
    cp = [i for i, x in enumerate(_line) if x == ")"]
    commas = [i for i, x in enumerate(_line) if x == ","]
    splitted_line = [char for char in _line]
    for n in commas:
        if len([x for x in op if x < n]) == len([x for x in cp if x < n]):
            splitted_line[n] = "#"
    _line = "".join(splitted_line)
    return [_x.strip() for _x in _line.split("#")]


def add_val_to_argument(string, _field):
    num_arguments, intervals = find_argument_intervals(string)
    for i in [_field]:
        num_arguments, intervals = find_argument_intervals(string)
        interval = intervals[i]
        start = interval[0]
        end = interval[1]
        if string[start] == "/":
            start += 1
        if string[end - 1] == "/":
            end -= 1
        first_segment = string[:start]
        intermediate_segment = string[start:end]
        last_segment = string[end:]
        string = "%s %s%%val %s" % (first_segment, intermediate_segment, last_segment)
    return string


def in_which_block(_line, _routine, vault=None):
    # Routine to know if we are entering or exiting a routine
    # If the line corresponds to the start of a routine, returns the name of this routine
    # If the line corresponds to the end of a routine, returns None
    # If is neither one of these cases, returns the same routine.
    _line = _line.strip().lower()
    _line = remove_literals(_line)
    if _line.count("subroutine"):
        if _line.strip().lower().find("end") == 0:
            return None
        else:
            match = re.search("subroutine *(\w*) *\(*", _line, flags=re.IGNORECASE)
            block_name = match.groups(0)[0].strip()
            if vault:
                procedure = vault.get_procedure(block_name)
                return procedure
            else:
                return block_name
    elif _line.count("function"):
        if _line.find("end") == 0:
            return None
        else:
            match = re.search("function *(\w+) *\(*", _line, flags=re.IGNORECASE)
            block_name = match.groups(0)[0].strip()
            if vault:
                procedure = vault.get_procedure(block_name)
                return procedure
            else:
                return block_name
    return _routine


def remove_comments(_text):
    # Trying to use a rarely used character ({) for the temporal replacement
    if _text.count("!"):
        # In the case where there's quotes in the text, we should ignore the exclamations inside
        if not _text.count("'"):
            _text = _text[:_text.find("!")]
        else:
            # Looking for
            scan_quotation = [_text[:i].count("'") for i, c in enumerate(_text)]
            splitted_text = [c for c in _text]
            splitted_text = ["{" if c == "!" and scan_quotation[i] % 2 == 1 else c for i, c in enumerate(splitted_text)]
            _text = "".join(splitted_text)

            if _text.find("!") >= 0:
                _text = _text[:_text.find("!")]
            _text = _text.replace("{", "!")

    return _text


def clean_argument(argument):
    # hashed_var_types, var_types
    #
    # derived_types = load_derived_type_info()
    # Returns the string removing the brackets and the optional dummy argument spec.

    if argument.count("%"):
        argument = argument.split("%")[-1].strip()
    else:
        if argument.count("("):
            argument = argument[:argument.find("(")].strip()
        if argument.count("="):
            argument = argument.split("=")[-1].strip()
        argument = argument.strip()
    return argument


def find_argument_intervals(_line, _field=None):
    # Given a line with a call to a function, give the index intervals of the different arguments.

    # Find all opening parenthesis
    op = [i for i, x in enumerate(_line) if x == "("]
    # Find all closing parenthesis
    cp = [i for i, x in enumerate(_line) if x == ")"]

    if not op and not cp:
        return 0, None

    # Keep only arguments part
    arguments_text = _line[op[0] + 1:cp[-1]]
    # Find remaining parenthesis
    start = op[0] + 1
    end = cp[-1]
    op = [i for i, x in enumerate(arguments_text) if x == "("]
    cp = [i for i, x in enumerate(arguments_text) if x == ")"]
    # Find commas

    splitted_line = [c for c in arguments_text]
    lit_counter = 0
    commas = []
    for i, c in enumerate(splitted_line):
        # Had to add this to avoid problems with commas inside strings"
        if c in ["'", '"']:
            lit_counter += 1
        if c == ",":
            if lit_counter % 2 == 0:
                commas.append(i)

    # For existing commas:
    argument_separating_commas = []
    for n in commas:
        # If comma index n is outside parenthesis (number of openings equal number of closings) substitute
        if len([x for x in op if x < n]) == len([x for x in cp if x < n]):
            splitted_line[n] = "#"
            argument_separating_commas.append(n)

    # Displacing the argument_separating_commas to match the original line
    argument_separating_commas = [_x + start + 1 for _x in argument_separating_commas]
    # Adding first and last elements
    argument_separating_commas = [start] + argument_separating_commas + [end + 1]

    # Constructing intervals
    intervals = [(argument_separating_commas[i], argument_separating_commas[i + 1] - 1) for i in
                 range(len(argument_separating_commas) - 1)]
    if _field is not None:
        if len(intervals) <= _field:
            return len(intervals), None
        else:
            return len(intervals), intervals[_field]
    else:
        return len(intervals), intervals


def is_keyword_argument(argument):
    pattern = "^[^()']*="
    if re.search(pattern, argument):
        return True
    else:
        return False


def add_function_to_argument(string, field_index, _function, number_of_fields=None):
    # Given a line with a call, add a call to a function in a specific field by index.
    num_arguments, interval = find_argument_intervals(string, field_index)
    if num_arguments <= field_index:
        return string
    if number_of_fields:
        if number_of_fields is not num_arguments:
            return string
    first_segment = string[:interval[0]]
    intermediate_segment = string[interval[0]:interval[1]]
    last_segment = string[interval[1]:]

    # The arguments containing a literal are discarded (Probably a bad workaround)
    if intermediate_segment.count("'"):
        # return _line
        raise ValueError("Probably the argument should not contain quotes at this point")

    # The arguments that already contain a call to the function are discarded (Probably a bad workaround)
    if re.search("%s *\(" % _function, intermediate_segment):
        # raise ValueError("Probably the argument should not contain %s at this point" % _function)
        return string

    # If there are named arguments, add the function to the actual argument
    # (if the equal sign is inside a parenthesis, this should not be done)
    named_argument_workaround = False

    if intermediate_segment.count("="):
        if intermediate_segment.count("("):
            if intermediate_segment.find("=") < intermediate_segment.find("("):
                named_argument_workaround = True
        else:
            named_argument_workaround = True

    if named_argument_workaround:
        parts = intermediate_segment.split("=")
        intermediate_segment = "%s=%s( %s )" % (parts[0], _function, parts[1])
    else:
        intermediate_segment = "%s( %s )" % (_function, intermediate_segment)
    string = "%s%s%s" % (first_segment, intermediate_segment, last_segment)
    return string


def add_rpe_to_argument(string, field_index, number_of_fields=None):
    string = add_function_to_argument(string, field_index, "rpe", number_of_fields)
    return string


def add_rpe_var_to_arguments(string):
    num_arguments, intervals = find_argument_intervals(string)
    for i in range(num_arguments):
        num_arguments, intervals = find_argument_intervals(string)
        interval = intervals[i]
        start = interval[0]
        end = interval[1]
        if string[start] == "/":
            start += 1
        if string[end - 1] == "/":
            end -= 1
        first_segment = string[:start]
        intermediate_segment = string[start:end]
        last_segment = string[end:]
        string = "%s rpe_var( 52, %s ) %s" % (first_segment, intermediate_segment, last_segment)
    return string


def add_real_to_argument(string, field_index, number_of_fields=None):
    string = add_function_to_argument(string, field_index, "real", number_of_fields)
    return string


def find_kind_of_parameter_exception(arguments):
    if not arguments.count(","):
        type_of_exception = "regular"
    else:
        if arguments.count("=") > 1:
            type_of_exception = "multiple_variables"
        elif arguments.count("(/"):
            type_of_exception = "array_initialization"
        elif arguments.count("rpe_var("):
            type_of_exception = "already_fixed"
        else:
            raise TypeError("Unexpected kind of exception")
    return type_of_exception


def remove_brackets_content(string):
    op = [i for i, x in enumerate(string) if x == "("]
    cp = [i for i, x in enumerate(string) if x == ")"]
    _new_line = []
    for n in range(len(string)):
        # If comma index n is outside parenthesis (number of openings equal number of closings) substitute
        if len([x for x in op if x <= n]) == len([x for x in cp if x < n]):
            _new_line += [string[n]]
    return "".join(_new_line)


def split_write_arguments(string):
    # Given a line containing a call to a write function, it returns the list of the arguments

    implicit_for_pattern = "WRITE(.*, *\w* *=[^,]*, *[\w0-9-+]* *\))"
    matches = re.search(implicit_for_pattern, string, re.IGNORECASE)
    if matches:
        implicit_for_part = "".join(
            list(reversed(close_inverse_brackets("".join(list(reversed(matches.groups()[0])))))))
        return []

    implicit_for_pattern_with_step = "WRITE(.*, *\w* *=[^,]*, *[\w0-9-+]* *, *[\w0-9-+]* *\))"
    matches = re.search(implicit_for_pattern_with_step, string, re.IGNORECASE)
    if matches:
        implicit_for_part = "".join(
            list(reversed(close_inverse_brackets("".join(list(reversed(matches.groups()[0])))))))

        return []

    pattern = "^[^\!]* +write\([^()]*\) *(.*)"
    matches = re.search(pattern, string, re.IGNORECASE)
    arguments = matches.groups(0)[0].strip()

    if not arguments.strip():
        return []

    has_brackets = False
    if all_is_inside_parenthesis(arguments):
        # if arguments[0] == "(" and arguments[-1] == ")":
        arguments = arguments[1:-1]
        has_brackets = True

    scan_quotation = [arguments[:i].count("'") for i, c in enumerate(arguments)]
    scan_op_brackets = [arguments[:i].count("(") for i, c in enumerate(arguments)]
    scan_cl_brackets = [arguments[:i].count(")") for i, c in enumerate(arguments)]
    scan_brackets = [i - j for i, j in zip(scan_op_brackets, scan_cl_brackets)]

    splitted_text = [c for c in arguments]
    splitted_text = ["#" if c == "," and scan_quotation[i] % 2 == 0 and not scan_brackets[i] else c for i, c in
                     enumerate(splitted_text)]
    arguments = "".join(splitted_text)
    arguments = [a for a in arguments.split("#") if a.strip()]
    if has_brackets:
        # TODO: I suspect that this can carry problems
        pass
        # arguments = ["(%s" % arguments[0]]+arguments[1:-1] + ["%s)"% arguments[-1]]
    return arguments


def split_read_arguments(string):
    # Given a line containing a call to a write function, it returns the list of the arguments
    # pattern = "^[^\!]* +read\([^()]*\) *(.*)"
    # matches = re.search(pattern, string, re.IGNORECASE)
    arguments = string

    scan_quotation = [arguments[:i].count("'") for i, c in enumerate(arguments)]
    scan_op_brackets = [arguments[:i].count("(") for i, c in enumerate(arguments)]
    scan_cl_brackets = [arguments[:i].count(")") for i, c in enumerate(arguments)]
    scan_brackets = [i - j for i, j in zip(scan_op_brackets, scan_cl_brackets)]
    splitted_text = [c for c in arguments]
    splitted_text = ["#" if c == "," and scan_quotation[i] % 2 == 0 and not scan_brackets[i] else c for i, c in
                     enumerate(splitted_text)]
    arguments = "".join(splitted_text)
    arguments = [a for a in arguments.split("#") if a.strip()]
    return arguments


def resulting_type(type_a, type_b):
    types = ["rpe_var", "complex", "real", "integer", "char", "logical"]
    for t in types:
        if type_a == t:
            return t
        if type_b == t:
            return t


def combine_types(list_of_types):
    to_return = list_of_types[0]
    for t in list_of_types:
        to_return = resulting_type(to_return, t)
    return to_return


def calls_to_routines(string):
    pattern = "[ \t]call +"
    matches = [(m.start(0), m.end(0)) for m in re.finditer(pattern, string, flags=re.I)]
    if matches:
        pieces = [string[start:].strip() for start, end in matches]
        return [close_brackets(piece) for piece in pieces if piece.count("(")]
    else:
        return None


def remove_literals(string):
    literal_map = [string[:i].count("'") % 2 for i, c in enumerate(string)]
    modified_string = "".join([c if (not literal_map[i] or c == "'") else " " for i, c in enumerate(string)])
    return modified_string


def calls_to_functions(string, vault=None):
    return calls_to_procedure(string, procedure_type=Function, vault=vault)


def calls_to_procedure(string, procedure_type, vault=None):
    # line without literals to avoid parenthesis inside strings and other ugly stuff
    no_literal_line = remove_literals(string)
    to_return = []
    if re.search(".*function .*", string, re.IGNORECASE):
        return to_return
    if re.search(".*subroutine.*", string, re.IGNORECASE):
        return to_return
    pattern = "(\w+) *\("
    matches = [m for m in re.finditer(pattern, no_literal_line, flags=re.I)]
    if matches:
        if vault:
            for m in matches:
                fun_name = m.groups(0)[0]
                fun = vault.get_procedure(fun_name, procedure_type)
                # TODO: Think a way to avoid having nf90 exceptions inside routines
                if fun or (procedure_type == Function and fun_name.lower().find("nf90_") == 0):
                    function_call = close_brackets(string[m.start(1):])
                    to_return.append(function_call)
        else:
            for m in matches:
                function_call = close_brackets(string[m.start(1):])
                to_return.append(function_call)

    return to_return


def find_called_routine(string):
    pattern = "^[^\!']*[\t ]*call +([^\(]*)"
    match = re.search(pattern, string, flags=re.I)
    if match:
        return match.groups()[0].strip().lower()
    else:
        return None


def find_called_function(string):
    return string[:string.find("(")].strip()


def find_call_arguments(string):
    n_arg, ints = find_argument_intervals(string)
    if n_arg == 0:
        return None
    if ints:
        arguments = [string[interval[0]:interval[1]].strip() for interval in ints]
        return [x for x in arguments if x]
    else:
        return None


def all_is_inside_parenthesis(string):
    if string.strip()[0] == "(":
        try:
            if "c%s" % string.strip() == close_brackets("c%s" % string.strip()):
                return True
            else:
                return False
        except ValueError:
            return False
    else:
        return False


def clean_contents(contents, vault=None):
    if vault:
        intrinsics = vault.intrinsics
    else:
        intrinsics = ["reshape", "min", "max", "sum", "log", "maxval", "minval", "abs", "size", "nint"]

    if re.match("\(/(.*)/\)", contents):
        contents = re.match("\(/(.*)/\)", contents).groups(1)[0]

    if re.match("^ *\(+ *(.*)", contents):
        contents = re.match("^ *\(+ *(.*)", contents).groups(1)[0]

    if re.match("^([^\()]*) *\)+ *$", contents):
        contents = re.match("^([^\()]*) *\)+ *$", contents).groups(1)[0]

    if re.match(".*\) *\(.*\)", contents):
        match = re.match("(.*\)) *\(.*\)", contents)
        contents = match.groups(0)[0]

    pattern = "(\w+) *\("
    m = [_m for _m in re.finditer(pattern, contents)]
    if m:
        for mat in reversed(m):
            prefix = contents[:mat.start(1)]
            part_to_modify = contents[mat.start(1):]
            to_replace = close_brackets(part_to_modify)
            replacement = to_replace[:to_replace.find("(")]
            if replacement in intrinsics:
                continue
            contents = "%s%s" % (prefix, part_to_modify.replace(to_replace, replacement))
    if contents.count("(/"):
        # Is temporal array
        contents = contents.replace("(/", "")
        contents = contents.replace("/)", "")
    return contents


def is_integer_with_precision(string):
    pattern = "^[0-9]*_\w+$"
    if re.search(pattern, string):
        return True
    else:
        return False


def is_call_to_function(contents):
    pattern = "\w+ *\("
    m = re.match(pattern, contents)
    if m:
        if contents == close_brackets(contents):
            return True
    return False


def type_of_contents(contents, routine=None, vault=None):
    # TODO: Simplify this function
    # I should do a first filter at this point distinguishing if its a string, a simple variable ("\w+")
    # if its a function/array (is_call_to_function is True), is a keyword argument, or other exceptions
    if not vault:
        vault = load_vault()

    # if all_is_inside_parenthesis(contents):
    #     contents = contents.strip()[1:-1]

    # Case with call to a function with array arguments after
    match = re.search("(\w* *\([^=]*\)) *(\(.*\))", contents)
    if match:
        contents = match.group(1)

    if contents.strip()[0] in ["'", '"']:
        return "char"

    comparison_operators = ["<", "==", ">"]
    for comparison_operator in comparison_operators:
        if contents.count(comparison_operator):
            if len(split_elements_in_comparison(contents)) > 1:
                return "logical"

    contents = contents.replace(" ", "").lower().strip()

    if is_call_to_function(contents):
        fun_name = find_called_function(contents)
        if fun_name in vault.intrinsics:
            if fun_name in ["reshape", "min", "max", "sum", "log", "maxval", "minval", "abs", "sqrt"]:
                return type_of_contents(find_call_arguments(contents)[0], routine=routine, vault=vault)
            elif fun_name in ["int", "nint", "count", "size", "mod", "umask", "len"]:
                return "integer"
            elif fun_name in ["real", "float"]:
                return "real"
            elif fun_name in ["cmplx"]:
                return "complex"
            elif fun_name in ["trim", "ctime", "adjustl"]:
                return "char"
            elif fun_name in ["null"]:
                return "pointer"

            else:
                raise ValueError("This intrinsic should be considered too: %s" % fun_name)
        else:
            fun = vault.get_function(fun_name)
            if fun:
                return fun.type
            fun = vault.get_procedure(fun_name, Interface)
            if fun:
                return fun.procedures[0].type

            # warnings.warn("If its a function should be on the library or an intrinsic, isn't it? %s|nLooking if its a variable... " % contents)
            # return type_of_contents(fun_name, routine=routine, vault=vault)

    contents = clean_contents(contents, vault)

    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        return type_of_contents(contents, routine=routine, vault=vault)

    if has_operations(contents):
        split_elements = split_elements_in_operations(contents)
        types = [type_of_contents(element, routine, vault) for element in split_elements]
        return combine_types(types)
    else:
        # if contents.find("size(") == 0:
        #     return "integer"
        # if contents.find("int(") == 0:
        #     return "integer"
        #
        # if contents.find("nint(") == 0:
        #     return "integer"
        # if contents.find("trim") == 0:
        #     return "char"
        # if contents.find("cmplx") == 0:
        #     return "complex"
        # if contents.find("real") == 0:
        #     return "real"

        if re.search("^[0-9\.]*e$", contents):
            return "real"

        if contents.find("nf90_") == 0:
            return "integer"
        if contents.find("set_nf90_") == 0:
            return "integer"

        if contents.count("'") or contents.count('"'):
            return "char"
        if contents.count(","):
            # If there are commas at this point, it must correspond with a temporal array
            # Array elements should have the same type, so we will assume that all the elements
            # have the same type than the first element
            first_element = split_attributes(contents)[0]
            return type_of_contents(first_element, vault=vault)
        if contents.count(","):
            raise ValueError("Argument should not contain commas at this point %s" % contents)

        if contents.count("(") or contents.count(")"):
            warnings.warn("WARNING: Argument should not contain brackets at this point % s" % contents)
            contents = contents.replace("(", "")
            contents = contents.replace(")", "")

        if contents.count(".false.") or contents.count(".true."):
            return "logical"
        if contents.count("."):
            return "real"
        if contents.isdigit():
            return "integer"
        if is_integer_with_precision(contents):
            return "integer"

        if vault:
            if re.search("^\w+$", contents):
                coincidences = [v for v in vault.get_variable(contents)]
                if len(coincidences) == 1:
                    return coincidences[0].type
                elif len(coincidences) > 1:
                    ctypes = [v.type for v in coincidences]
                    if elements_are_equal(ctypes):
                        return ctypes[0]

            if routine:
                if isinstance(routine, Routine) or isinstance(routine, ModuleMain) or isinstance(routine, Function):
                    procedure = routine
                else:
                    procedure = vault.get_procedure(routine)

                for variable in procedure.variables:
                    if contents == variable.name.lower():
                        return variable.type
                for variable in procedure.accessible_variables():
                    if contents == variable.name.lower():
                        return variable.type

                # Try if there's a function with this name
                fun = vault.get_function(contents)
                if fun:
                    return fun.type

        if contents.count("%"):
            return type_of_derived_type_element(contents, routine=routine, vault=vault)

        if contents.count(">") or contents.count("<") or contents.count("=="):
            return "logical"

        if vault:
            if contents.count("mpi_"):
                return "integer"
            elif contents == "rpe":
                return "rpe_var"
            else:
                warnings.warn("Not able to find type! :(  |%s|" % contents)
        else:
            pass

        warnings.warn("Did not find type of: %s" % contents)
        # raise AssertionError("Did not find type of: %s" % contents)


def retrive_variable(contents, routine=None, vault=None):
    # TODO: Simplify this function
    # I should do a first filter at this point distinguishing if its a string, a simple variable ("\w+")
    # if its a function/array (is_call_to_function is True), is a keyword argument, or other exceptions
    if not vault:
        vault = load_vault()

    contents = contents.replace(" ", "").lower().strip()

    contents = clean_contents(contents, vault)

    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        return retrive_variable(contents, routine=routine, vault=vault)

    if vault:
        if re.search("^\w+$", contents):
            coincidences = [v for v in vault.get_variable(contents)]
            if len(coincidences) == 1:
                return coincidences[0]

        if routine:
            if isinstance(routine, Routine) or isinstance(routine, ModuleMain) or isinstance(routine, Function):
                procedure = routine
            else:
                procedure = vault.get_procedure(routine)

            for variable in procedure.variables:
                if contents == variable.name.lower():
                    return variable
            for variable in procedure.accessible_variables():
                if contents == variable.name.lower():
                    return variable

            # Try if there's a function with this name
    if contents.count("%"):
        contents = contents.split("%")[0]
        return retrive_variable(contents, routine=routine, vault=vault)

    # warnings.warn("Did not find type of: %s" % contents)
    raise AssertionError("Did not find this variable: %s" % contents)


def type_of_derived_type_element(contents, routine=None, derived_type=None, vault=None):
    if not vault:
        vault = load_vault()

    struct = contents[:contents.find("%")]
    var = contents[contents.find("%") + 1:]
    if var == "val":
        return "real"

    if not derived_type:
        derived_type = type_of_contents(struct, routine=routine, vault=vault)

    if var.count("%"):

        struct = var[:var.find("%")]
        derived_type = type_of_derived_type_element(struct, routine=routine, derived_type=derived_type, vault=vault)
        var = var[var.find("%") + 1:]
        if var == "val":
            return "real"
        dt = vault.get_derived_type(derived_type)
        try:
            v = dt.get_variable(var)
        except AttributeError:
            pass

        try:
            return v.type
        except AttributeError:
            pass
    else:
        if vault:
            dt = vault.get_derived_type(derived_type)
            if not dt:
                pass
            try:
                return dt.get_variable(var).type
            except AttributeError:
                pass

        return "unknown"


def load_pickle_file(filename):
    try:
        with open(filename, "rb") as _f:
            data = pickle.load(_f)
    except IOError:
        warnings.warn("File %s not found, not using rotuines info." % filename)
        data = None
    return data


def load_source_files(path):
    filelist = [ReadSourceFile(join(path, filename)) for filename in listdir(path)]
    filedict = {}
    for f in filelist:
        filedict[f.filename] = f
    return filedict


# def load_unified_dictionary():
#     return load_pickle_file("unified_dictionary.pkl")
#
#
# def load_routine_info():
#     # Dictionary with routine arguments
#     _routines = load_pickle_file("routines.pkl")
#     _external_routines = load_pickle_file("external_routines.pkl")
#     return _routines, _external_routines
#
#
# def load_functions_info():
#     return load_pickle_file("functions.pkl")
#
#
# def load_derived_type_info():
#     return load_pickle_file("derived_types.pkl")
#
#
# def load_variable_info():
#     _hashed_vars = load_pickle_file("hashed_var_types.pkl")
#     _vars = load_pickle_file("var_types.pkl")
#     return _hashed_vars, _vars


def load_vault():
    try:
        return pickle.load(open("vault.pkl", "rb"))
    except IOError:
        return None


def load_intrinsics():
    # Hardcoded path to file.
    with open("AdditionalFiles/fortran_intrinsics") as f:
        return [re.search("(\w+)", l).groups(0)[0].lower() for l in f]


def load_list_of_files_to_keep_unmodified():
    # Hardcoded path to file.
    filename = "AdditionalFiles/list_of_files_to_keep_unmodified.txt"
    if isfile(filename):
        print("Loading list of exceptions from %s\n\n" % filename )
        with open("AdditionalFiles/list_of_files_to_keep_unmodified.txt") as f:
            return [l.strip() for l in f if l.strip() and l.strip()[0] != "#"]
    else:
        return None


def close_brackets(string):
    started = False
    bracket_counter = 0
    literal_counter = 0
    for _i, _c in enumerate(string):
        if _c == "'":
            literal_counter += 1
        if _c == "(":
            if not started:
                started = True
            if not literal_counter % 2:
                bracket_counter += 1
        elif _c == ")":
            if not started:
                raise ValueError("This should not happen, check arguments %s" % string)
            if not literal_counter % 2:
                bracket_counter -= 1
        if bracket_counter < 0:
            raise ValueError("This should not happen, check arguments %s" % string)
        if started and not bracket_counter:
            return string[:_i + 1]

    if string.count("(") + string.count(")") == 0:
        return string
    raise ValueError("This should not happen, check arguments")


def close_inverse_brackets(string):
    started = False
    bracket_counter = 0
    literal_counter = 0
    for _i, _c in enumerate(string):
        if _c == "'":
            literal_counter += 1
        if _c == ")":
            if not started:
                started = True
            if not literal_counter % 2:
                bracket_counter += 1
        elif _c == "(":
            if not started:
                raise ValueError("This should not happen, check arguments %s" % string)
            if not literal_counter % 2:
                bracket_counter -= 1
        if bracket_counter < 0:
            raise ValueError("This should not happen, check arguments %s" % string)
        if started and not bracket_counter:
            return string[:_i + 1]

    if string.count("(") + string.count(")") == 0:
        return string
    raise ValueError("This should not happen, check arguments")


def elements_are_equal(iterator):
    iterator = iter(iterator)
    try:
        first = next(iterator)
    except StopIteration:
        return True
    return all(first == rest for rest in iterator)


def split_declaration(declaration):
    # Description: Given a declaration line, this function gives back the attributes and the declared arguments
    # Input:
    #       -declaration: string corresponding to a variable declaration
    # Output:
    #       _attributes: string corresponding to the declaration attributes
    #       _arguments: string containing the declaration arguments
    _attributes, _arguments = declaration.split("::")
    _attributes = _attributes.strip()
    _arguments = split_declaration_arguments(_arguments)
    return _attributes, _arguments


def split_attributes(attributes):
    # Description: Given a string with declaration attributes separates the different attributes and
    # gives back a list of attributes
    # Input:
    #       -attributes: string with attributes
    # Output:
    #       -attributes_list: list containg strings with the different attributes
    op = [i for i, x in enumerate(attributes) if x == "("]
    cp = [i for i, x in enumerate(attributes) if x == ")"]
    tmp_attributes = [c for c in attributes]
    for n in range(len(attributes)):
        # If comma index n is outside parenthesis (number of openings equal number of closings) substitute
        if attributes[n] == ",":
            if len([x for x in op if x <= n]) != len([x for x in cp if x < n]):
                tmp_attributes[n] = "#"
    attributes_list = "".join(tmp_attributes)
    attributes_list = [l.strip().replace("#", ",") for l in attributes_list.split(",")]
    return attributes_list


def which_type(attributes):
    # Description: Given a list of attributes, this function returns the type of the variables
    # Input: List of attributes
    # Output: String with type
    if isinstance(attributes, str):
        attributes = split_attributes(attributes)
    assert isinstance(attributes, list)
    possible_types = ["real", "integer", "type", "char", "logical", "complex", "double precision", "useintrinsic",
                      "public"]
    t0 = attributes[0]
    var_type = None
    for ty in possible_types:
        if t0.lower().find(ty) == 0:
            var_type = ty
    if not var_type:
        return None
    if var_type == "type":
        # It may be the case in which it is a type declaration instead of a variable declaration of a type
        # This workaround is to discard this cases.
        if not t0.count("("):
            return None
        var_type = t0[t0.find("(") + 1:-1].strip()
    if var_type == "public":
        return None

    return var_type.lower()


def which_dimension(attributes):
    # Description: Given a list of attributes, this function returns the dimension of the variables
    # Input: List of attributes
    # Output: Integer with the number of dimensions
    attributes = split_attributes(attributes)
    dim = 0
    for att in attributes:
        if att.lower().count("dimension"):
            dim = att
            break
    if dim:
        dim = 1 + dim.count(",")
    return dim


def which_intent(attributes):
    # Description: Given a list of attributes, this function returns the intent of the variables
    # Input: List of attributes
    # Output: String with the intent of the variable, if intent is not specified the default is inout
    attributes = split_attributes(attributes)
    intent = "inout"
    for att in attributes:
        if att.lower().count("intent"):
            intent = att
            break
    if intent.count("("):
        intent = intent[intent.find("(") + 1:-1].strip().lower()
    return intent


def is_optional(attributes):
    # Description: Given a list of attributes, this function returns a boolean describing if the variable is optional
    # Input: List of attributes
    # Output: Boolean about optionality of variable
    attributes = split_attributes(attributes)
    optional = False
    for att in attributes:
        if att.lower().count("optional"):
            optional = True
            break
    return optional


def is_parameter(attributes):
    # Description: Given a list of attributes, this function returns a boolean describing if the variable is optional
    # Input: List of attributes
    # Output: Boolean about optionality of variable
    attributes = split_attributes(attributes)
    for att in attributes:
        if att.lower().count("parameter"):
            return True
    return False


def is_allocatable(attributes):
    # Description: Given a list of attributes, this function returns a boolean describing if the variable is allocatable
    # Input: List of attributes
    # Output: Boolean about if variable is allocatable
    attributes = split_attributes(attributes)
    for att in attributes:
        if att.lower().count("allocatable"):
            return True
    return False


def parse_attributes(attributes):
    if isinstance(attributes, str):
        attributes = split_attributes(attributes)
    assert isinstance(attributes, list)
    var_type = which_type(attributes)
    dimension = which_dimension(attributes)
    intent = which_intent(attributes)
    optional = is_optional(attributes)
    allocatable = is_allocatable(attributes)
    parameter = is_parameter(attributes)
    return var_type, dimension, intent, optional, allocatable, parameter


def split_declaration_arguments(string):
    # Description: given a string containing arguments, return a list with arguments
    # Input: String with arguments
    # Output: List with arguments
    # TODO: If the string of arguments contains parenthesis, we should keep information about the dimensions.
    # TODO: If the argument has a value declaration, we should get rid of this.
    op = [i for i, x in enumerate(string) if x == "("]
    cp = [i for i, x in enumerate(string) if x == ")"]
    line = [c for c in string]
    for n in range(len(string)):
        # If comma index n is outside parenthesis (number of openings equal number of closings) substitute
        if string[n] == ",":
            if len([x for x in op if x <= n]) != len([x for x in cp if x < n]):
                line[n] = "#"
    string = "".join(line)
    string = [l.strip().replace("#", ",") for l in string.split(",")]
    string = [el[:el.find("(")] if el.count("(") else el for el in string]
    string = [el[:el.find("=")].strip() if el.count("=") else el for el in string]
    return string


def split_elements_in_operations(string):
    operations = ["+", "-", "*", "/"]

    functions = calls_to_functions(string)
    dictionary_of_replacements = {}
    if functions:
        for index, function in enumerate(functions):
            dictionary_of_replacements["REPLACEMENT%i" % index] = function
            string = string.replace(function, "REPLACEMENT%i" % index)

    listed_text = [c for c in string]
    for index, c in enumerate(listed_text):
        if c in operations:
            listed_text[index] = "#"
    string = "".join(listed_text)
    arguments = [x.strip() for x in string.split("#") if x.strip()]
    for index, argument in enumerate(arguments):
        for key, function in dictionary_of_replacements.items():
            if key in argument:
                arguments[index] = arguments[index].replace(key, function)

    return arguments


def split_elements_in_comparison(string):
    comparison_operators = [">", "<", "=="]

    functions = calls_to_functions(string)
    dictionary_of_replacements = {}
    if functions:
        for index, function in enumerate(functions):
            dictionary_of_replacements["REPLACEMENT%i" % index] = function
            string = string.replace(function, "REPLACEMENT%i" % index)

    listed_text = [c for c in string]
    for index, c in enumerate(listed_text):
        if c in comparison_operators:
            listed_text[index] = "#"
    string = "".join(listed_text)
    arguments = [x.strip() for x in string.split("#") if x.strip()]
    for index, argument in enumerate(arguments):
        for key, function in dictionary_of_replacements.items():
            if key in argument:
                arguments[index] = arguments[index].replace(key, function)

    return arguments


def has_operations(string):
    operations = ["+", "-", "*", "/"]
    if not string.strip():
        return False
    if string.count("'") or string.count('"'):
        return False
    # TODO: Does it break something else?
    # if string.count("-"):
    #     if string.count("-") == string.count("e-"):
    #         return False
    for op in operations:
        if op in string:
            return True
    return False


def match_arguments_and_interface(arguments, interface, current_block=None, vault=None):
    number_of_arguments = len(arguments)
    argument_types = [type_of_contents(argument, routine=current_block, vault=vault) for argument in arguments]
    keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in arguments]
    for procedure in interface.procedures:
        dummy_arguments = procedure.get_dummy_arguments()
        mandatory_arguments = [argument for argument in dummy_arguments if not argument.is_optional]
        if len(dummy_arguments) < number_of_arguments:
            continue
        if len(mandatory_arguments) > number_of_arguments:
            continue
        match = True
        for index, argument_type in enumerate(argument_types):
            keyword = keyword_arguments[index]
            if keyword:
                var = procedure.get_variable_by_name(keyword)
                if var:
                    dummy_type = var.type
                else:
                    match = False
            else:
                dummy_type = procedure.get_variable_by_position(index).type
            if dummy_type != argument_type:
                if (dummy_type == "real" and argument_type == "rpe_var") or (
                        dummy_type == "rpe_var" and argument_type == "real"):
                    pass
                else:
                    match = False
        if match:
            return procedure
    raise AssertionError("No matching routine found for interface : %s" % interface.name)


def function_type_from_header(header):
    pattern = "(.*)function"
    match = re.search(pattern, header, re.IGNORECASE)
    fun_type = remove_brackets_content(match.groups(0)[0].strip().lower())
    if fun_type:
        return True, fun_type
    else:
        if header.count("result"):
            pattern = "result *\((.*)\)"
            match = re.search(pattern, header, re.IGNORECASE)
            if match:
                return False, remove_brackets_content(match.groups(0)[0].strip().lower())
            else:
                raise ValueError("Not finding function result")

        else:
            pattern = "function (.*)\("
            match = re.search(pattern, header, re.IGNORECASE)
            if match:
                return False, remove_brackets_content(match.groups(0)[0].strip().lower())
            else:
                raise ValueError("Not finding function result")


def split_vars_by_allocatable(variables):
    allocatables = [var for var in variables if var.type == "rpe_var" and var.allocatable]
    non_allocatables = [var for var in variables if var.type == "rpe_var" and not var.allocatable]
    return allocatables, non_allocatables


def recover_routine_text(variable, file_objects):
    routine = variable.routine
    module = routine.module

    f = file_objects["%s.f90" % module.name]
    to_return = []
    lines = f.lines
    if routine.name == "main":
        for line in lines[:]:
            if line.lower().count("contains"):
                break
            else:
                to_return.append(line)
        to_return.append(lines[-1])
    else:
        printing = False
        for line in lines:
            if not printing:
                if line.lower().count("subroutine"):
                    called_routine = re.search("subroutine *(\w*)", line, re.IGNORECASE).group(1)
                    if called_routine.lower() == routine.name.lower():
                        to_return.append(line)
                        printing = True
            else:
                to_return.append(line)
                if line.lower().count("end subroutine"):
                    break

    return to_return


def is_allocated_inside_declaration_scope(var, file_objects):
    text = recover_routine_text(var, file_objects)
    interesting_lines = [line for line in text if
                         re.search("allocate *\(.*[^\w]%s[^\w]" % var.name, line, re.IGNORECASE)]
    if interesting_lines:
        return True
    else:
        return False


def split_by_function(variables, fun, files=None):
    pattern = "[^\w]%s[^\w]"

    vars_true = []
    vars_false = []

    for var in variables:
        appears_in = find_var_elsewhere(var, files)
        found = False
        for f in appears_in:
            appears_in_lines = [line for line in f.lines if
                                re.search(pattern % var.name, line) and not line.count("::")]
            found_array = [True if fun(line) else False for line in appears_in_lines]
            if any(found_array):
                found = True
        if found:
            vars_true.append(var)
        else:
            vars_false.append(var)
    return vars_true, vars_false


def is_namelist_line(line):
    return line.lower().strip().find("namelist") == 0


def is_assignation_line(line):
    return line.lower().strip().count("=") > 0


def split_by_appearance_in_namelist(variables, files=None):
    return split_by_function(variables, is_namelist_line, files=files)


def split_vars_by_main_or_routine(variables):
    in_main = [var for var in variables if var.routine.name == "main"]
    not_in_main = [var for var in variables if var.routine.name != "main"]

    return in_main, not_in_main


# Source Database Classes
class Vault:
    def __init__(self):
        self.modules = []
        self.variables = []
        self.dictionary_of_variables = {}
        self.dictionary_of_procedures = {}
        self.procedures = []
        self.routines = []
        self.functions = []
        self.interfaces = []
        self.derived_types = []
        self.intrinsics = load_intrinsics()

    def add_module(self, module):
        self.modules.append(module)

    def __getitem__(self, item):
        for module in self.modules:
            if module.name == item:
                return module

    def process_procedures(self):
        for module in self.modules:
            for procedure in module.procedures:
                self.procedures.append(procedure)
                # self.dictionary_of_procedures[procedure.name.lower()] = procedure
        for procedure in self.procedures:
            if isinstance(procedure, Routine):
                self.routines.append(procedure)
            elif isinstance(procedure, Function):
                self.functions.append(procedure)
            elif isinstance(procedure, Interface):
                self.interfaces.append(procedure)
            elif isinstance(procedure, DerivedType):
                self.derived_types.append(procedure)

    def process_variables(self):
        if not self.procedures:
            self.process_procedures()
        for procedure in self.procedures:
            if isinstance(procedure, ModuleMain) or isinstance(procedure, Function) or isinstance(procedure, Routine):
                for variable in procedure.variables:
                    self.variables.append(variable)
                    if not variable.name.lower() in self.dictionary_of_variables:
                        self.dictionary_of_variables[variable.name.lower()] = [variable]
                    else:
                        self.dictionary_of_variables[variable.name.lower()] += [variable]

    def get_variable(self, item):
        if not self.variables:
            self.process_variables()
        try:
            for variable in self.dictionary_of_variables[item.lower()]:
                yield variable
        except KeyError:
            pass

    def get_function(self, item):
        return self.get_procedure(item, Function)

    def get_routine(self, item):
        return self.get_procedure(item, Routine)

    def get_procedure(self, item, procedure_type=None):
        # TODO: This routine is quite expensive (16%), could be boosted using dictionaries.
        if not self.procedures:
            self.process_procedures()

        for procedure in self.procedures:
            if procedure_type:
                if isinstance(procedure, procedure_type):
                    if procedure.name == item:
                        return procedure
            elif procedure.name == item:
                return procedure

    def get_derived_type(self, item):
        for dt in self.derived_types:
            try:
                if dt.name.lower() == item.lower():
                    return dt
            except:
                pass

    def expand_used_modules(self):
        for module in self.modules:
            queue = []
            done = []
            for mod in module.uses:
                queue.append(mod)
            while len(queue) > 0:
                mod = queue.pop(0)
                for mod1 in mod.uses:
                    if mod1 not in module.uses and mod1 not in done:
                        module.uses.append(mod1)
                        queue.append(mod1)
                done.append(mod)

    def expand_used_by(self):
        for module in self.modules:
            for mod1 in module.uses:
                if module not in mod1.used_by:
                    mod1.used_by.append(module)

    def generate_namelist_precisions(self, filename="namelist_precisions", exclude_double_precision=True):
        rpe_vars = [var for var in self.variables if var.type == "rpe_var"]
        assigned_rpe_vars = [var for var in rpe_vars if var.id is not None]
        assigned_rpe_vars = sorted(assigned_rpe_vars, key=lambda x: x.id)
        line_text = "emulator_variable_precisions(%d) = %d\t! Variable:%24s\tRoutine:%22s\tModule:%16s\n"
        with open(filename, "w") as namelist:
            namelist.write("! namelist variable precisions\n&precisions\n")
            for var in assigned_rpe_vars:
                if var.precision == 52 and exclude_double_precision:
                    continue
                namelist.write(line_text % (var.id, var.precision, var.name, var.routine.name, var.routine.module.name))
            namelist.write("/")


class Module:
    def __init__(self, _name):
        self.name = _name
        self.repo = None
        self.procedures = []
        self.variables = []
        self.precision = None
        self.uses = []
        self.used_by = []

    def add_procedure(self, _routine):
        assert isinstance(_routine, Procedure)
        _routine.module = self
        self.procedures.append(_routine)

    def add_variable(self, _variable):
        assert isinstance(_variable, Variable)
        _variable.routine = self
        self.variables.append(_variable)

    def __getitem__(self, item):
        for _routine in self.procedures:
            if _routine.name == item:
                return _routine
        new_routine = Routine(None, item)
        new_routine.module = self
        self.repo.add_procedure(new_routine)
        self.procedures.append(new_routine)
        return self[item]

    def set_precision(self, precision):
        assert isinstance(precision, int), "Precision argument in set_precision must be an integer"
        for routine in self.procedures:
            routine.set_precision(precision)
        self.precision = precision
        self.repo.list_of_precision_is_updated = False

    def routine_names(self):
        return [el.name for el in self.procedures]

    def main(self):
        for procedure in self.procedures:
            if isinstance(procedure, ModuleMain):
                return procedure
        return None


class Procedure(object):
    def __init__(self, _name, module):
        self.name = _name
        self.module = module
        self.variables = []
        self.precision = None
        self.interface = None
        self.dummy_arguments = None
        self.uses = []

    def add_variable(self, _variable):
        try:
            assert isinstance(_variable, Variable)
        except AssertionError:
            pass
        _variable.routine = self
        self.variables.append(_variable)

    def get_variable_by_position(self, position):
        for variable in self.variables:
            if variable.is_dummy_argument:
                if variable.position == position:
                    return variable

    def get_variable_by_name(self, name):
        for variable in self.variables:
            if variable.name.lower() == name.lower():
                return variable

    def accessible_variables(self):
        variables = self.variables
        if self.module:
            variables += self.module.main().variables
            if self.module.uses:
                for module in self.module.uses:
                    variables += module.main().variables

        return variables

    def get_dummy_arguments(self):
        if self.dummy_arguments is None:
            number_of_dummy_arguments = len([True for var in self.variables if var.is_dummy_argument])
            self.dummy_arguments = [self.get_variable_by_position(i) for i in range(number_of_dummy_arguments)]
        return self.dummy_arguments


class Routine(Procedure):
    def __init__(self, _name, module):
        super(Routine, self).__init__(_name, module)


class Function(Procedure):
    def __init__(self, _name, module):
        super(Function, self).__init__(_name, module)
        self.type = None


class DerivedType(Procedure):
    def __init__(self, _name, module):
        super(DerivedType, self).__init__(_name, module)

    def get_variable(self, item):
        for variable in self.variables:
            if variable.name.lower() == item.lower():
                return variable


class ModuleMain(Procedure):
    def __init__(self, module):
        super(ModuleMain, self).__init__("main", module)
        self.module = module


class Interface:
    def __init__(self, _name):
        self.name = _name
        self.module = None
        self.procedures = []

    def add_procedure(self, _procedure):
        assert isinstance(_procedure, Routine) or isinstance(_procedure, Function)

        self.procedures.append(_procedure)
        # Don't know it will be useful yet
        if not _procedure.interface:
            _procedure.interface = self
        elif isinstance(_procedure.interface, list):
            _procedure.interface.append(self)
        else:
            _procedure.interface = [_procedure.interface, self]


class Variable:
    def __init__(self, _routine, _name, _type, _dimension):
        self.id = None
        self.module = None
        self.routine = _routine
        self.name = _name
        self.type = _type
        self.dimension = _dimension
        self.is_dummy_argument = False
        self.intent = None
        self.is_optional = None
        self.position = None
        self.allocatable = False
        self.input_of = []
        self.target_of = []
        self.interact_with_variables = []
        self.precision = None
        self.is_parameter = None

    def set_precision(self, precision):
        assert isinstance(precision, int), "Precision argument in set_precision must be an integer"
        self.precision = precision


# File manipulation Classes

class ReadSourceFile(object):
    def __init__(self, _filename):
        # Init variables
        self.file_path = _filename
        self.filename = self.file_path.split("/")[-1]
        self.text = None
        self.lines = None

        # Read file
        self.read_file(_filename)

        # Pre-Process file
        self.remove_all_comments()
        self.remove_empty_lines()
        self.merge_all()
        self.split_lines()
        self.split_declarations()
        self.rebuild_text()

    def read_file(self, _filename):
        with open(_filename) as f:
            self.text = f.read()
            self.lines = self.text.split("\n")

    def remove_comments_in_interval(self, start, end):
        # Definition: The comments of the lines between start and end are removed
        # Input:
        #       -start: integer with the starting line number.
        #       -end: integer with the ending line number.
        for _index in range(start, end):
            self.lines[_index] = remove_comments(self.lines[_index])

    def merge_all(self):
        self.merge_lines_in_interval(0, len(self.lines))

    def merge_lines_in_interval(self, start, end):
        # Definition: Given the number of the starting and ending lines of an interval,
        # the split lines are merged.
        # Input:
        #       -r_start: integer with the starting line number.
        #       -r_end: integer with the ending line number.
        for _index in range(start, end):
            line = self.lines[_index]
            if line.count("&"):
                i = 1
                while line and line.strip()[-1] == "&":
                    line1 = line
                    line2 = self.lines[_index + i]
                    line = line1.replace("&", "").rstrip() + line2

                    i += 1
                line = line.replace("&", "")
                line = re.sub(", +", ", ", line)
                line = re.sub(":  +", ":  ", line)
                self.lines[_index] = line
                for j in range(1, i):
                    self.lines[_index + j] = ""

    def split_lines(self):
        # TODO: We should be sure that we are not spliting by characters in literals

        _index = 0
        length = len(self.lines)
        while _index < length:
            line = self.lines[_index]
            if line.count(";"):
                parts = split_semicolon(line)
                self.lines[_index] = parts[0]
                for i in range(1, len(parts)):
                    self.lines.insert(_index + 1, parts[i])
                    length += 1
                    _index += 1
            _index += 1
        self.rebuild_text()

    def split_declarations(self):
        # This method should split declarations of variables putting a single variable in each line
        _index = 0
        length = len(self.lines)
        while _index < length:
            line = self.lines[_index]
            if line.count("::"):
                attributes, arguments = line.split("::")
                arguments = find_call_arguments("c(%s)" % arguments)
                if len(arguments) > 1:
                    new_line = "%s :: %s" % (attributes, arguments[0])
                    self.lines[_index] = new_line
                    for argument in arguments[1:]:
                        new_line = "%s :: %s" % (attributes, argument)
                        self.lines.insert(_index + 1, new_line)
                        length += 1
                        _index += 1

                # parts = split_semicolon(line)
            #         self.lines[_index] = parts[0]
            #         for i in range(1, len(parts)):
            #             self.lines.insert(_index+1, parts[i])
            #             length += 1
            #             _index += 1
            _index += 1
        self.rebuild_text()

    def remove_all_comments(self):
        self.remove_comments_in_interval(0, len(self.lines))
        self.remove_empty_lines()

    def remove_empty_lines(self):
        lines = [line for line in self.lines if line.strip()]
        self.lines = lines

    def rebuild_text(self):
        self.text = "".join([l + "\n" for l in self.lines[:]])

    def save_modified_file(self, filepath):
        with open(filepath, "w") as output_file:
            output_file.write(self.text)


# Classes used in ParseRoutines.py
class ObtainSourcefileInfo(ReadSourceFile):
    # Definition: This class is used to load a sourcefile and parse the subroutines contained there
    # Input: string with path to sourcefile
    def __init__(self, _filename):
        super(ObtainSourcefileInfo, self).__init__(_filename)
        # Definition: Initializing the object reading the sourcefile
        # Input: string with path to sourcefile
        # I think that it is save to collect the information using all the code in lower case.
        self.text = self.text.lower()
        self.objects = []
        self.used_modules = []
        self.module = None

    def find_next_routine_end(self, _index):
        # Definition: Given a line number, looks for next line with "end subroutine"
        # Input: integer with line number
        # Output: integer with the line number of the next "end subroutine"
        return self.find_next_end(_index, "end subroutine")

    def find_next_function_end(self, _index):
        # Definition: Given a line number, looks for next line with "end subroutine"
        # Input: integer with line number
        # Output: integer with the line number of the next "end subroutine"
        return self.find_next_end(_index, "end function")

    def find_next_interface_end(self, _index):
        # Definition: Given a line number, looks for next line with "end interface"
        # Input: integer with line number
        # Output: integer with the line number of the next "end interface"
        return self.find_next_end(_index, "end interface")

    def find_next_end(self, _index, end_statement):
        # Definition: Given a line number, looks for next line with the string contained in end_statement argument.
        # Input:
        #       _index: integer with line number
        #       end_statement: string that indicates next end_statement

        # Output: integer with the line number of the next line containing the end_statement
        _i = _index
        while _i < len(self.lines):
            _l = self.lines[_i]
            if _l.lower().count(end_statement):
                return _i
            _i += 1
        return len(self.lines)

    def get_dummy_arguments(self, _index):
        # Definition: Given a line number, returns the subroutine name and the dummy arguments (in case they exist)
        # Input: integer with line number
        # Output: tuple with routine name and list of dummy arguments
        _line = self.lines[_index]
        # Workaround to solve the problem that appears when there's a function declaration with (kind=...)
        if _line.lower().count("function"):
            if _line.find("(") < _line.lower().find("function"):
                _line = _line[_line.lower().find("function"):]

        # If the subroutine declaration does not have any parenthesis, this routine does not have any argument
        if not _line.count("("):
            _name = _line.split()[1].strip()
            _dummy_arguments = []
        # If otherwise it has dummy arguments, find them
        elif _line.lower().count("result") or _line.lower().count("bind("):
            _name = _line[:_line.find("(")].split()[1]
            _dummy_arguments = []
            _dummy_arguments += strip_call_arguments(close_brackets(_line))
            # I don't know if we should record the result variable with the other variables
            # result_start = _line.lower().find("result")
            # _dummy_arguments += strip_call_arguments(_line[result_start:])
        else:
            _name = _line[:_line.find("(")].split()[-1]
            _dummy_arguments = strip_call_arguments(_line)
        _dummy_arguments = [_x for _x in _dummy_arguments if _x.strip()]
        return _name, _dummy_arguments

    def find_routine_declarations(self, _routine_start, _routine_end):
        # Definition: Given the line numbers where a routine starts and ends,
        # look for variable declarations.
        # Input:
        #       -_routine_start: integer with line number where the routine starts
        #       -_routine_end: integer with line number where the routine ends
        # Output: list of strings containing declarations

        # Lines inside the routine
        _routine_lines = self.lines[_routine_start:_routine_end + 1]
        # Lines that are variable declarations
        _declarations = [_l for _l in _routine_lines if _l.count("::")]
        return _declarations

    def find_procedure_variables(self, routine_name, dummy_arguments, declarations, header):
        # Definition: Given the routine name, the name of the dummy arguments and the lines containing declarations,
        # this method finds the attributes of each dummy arguments and adds all the information
        # into the object variable "routines".
        # Input:
        #       -routine_name: string with the routine name
        #       -dummy_arguments: list of strings with argument names
        #       -declarations: list of strings with variable declarations

        if header.count("function"):
            routine_or_function = "function"
        elif header.count("subroutine"):
            routine_or_function = "routine"
        else:
            raise ValueError("Header does not correspond to function or routine")

        if routine_or_function == "routine":
            obj_type = Routine
        elif routine_or_function == "function":
            obj_type = Function
        else:
            raise ValueError("routine_or_function must be 'routine' or 'function'")

        ob = obj_type(routine_name, self.module)

        # Reverse approach
        dummy_arguments = [x.lower() for x in dummy_arguments]
        for declaration in declarations:
            _attributes, _arguments = split_declaration(declaration)
            # Check if the dummy argument is inside the declaration arguments
            # If it is, take the type, dimension, intent and optional
            for argument in _arguments:
                var_type = which_type(_attributes)
                dimension = which_dimension(_attributes)
                intent = which_intent(_attributes)
                # In case of functions, we have only one output variable and all the others are inputs
                if routine_or_function == "function":
                    if intent == "inout":
                        intent = "in"
                optional = is_optional(_attributes)
                allocatable = is_allocatable(_attributes)
                _v = Variable(ob, argument, var_type, dimension)
                _v.allocatable = allocatable
                _v.is_parameter = is_parameter(_attributes)

                if argument.lower() in dummy_arguments:
                    _v.intent = intent
                    _v.is_optional = optional
                    _v.is_dummy_argument = True
                ob.add_variable(_v)

        # Save the position of the argument
        found_vars = [variable.name.lower() for variable in ob.variables]
        for index, dummy_argument in enumerate(dummy_arguments):
            if dummy_argument in found_vars:
                # TODO: save position of already found vars
                ob.get_variable_by_name(dummy_argument).position = index  # variable.position = index
                continue
                # If it's not found its an implicitly declared variable
            if dummy_argument[0] in "ijklmn":
                var_type = "integer"
            else:
                var_type = "real"
            dimension = 0
            intent = None
            optional = None
            _v = Variable(ob, dummy_argument, var_type, dimension)
            _v.intent = intent
            _v.is_optional = optional
            _v.is_dummy_argument = True
            _v.position = index
            _v.is_parameter = False
            ob.add_variable(_v)
        # Save object
        if routine_or_function == "function":

            type_found, type_or_var = function_type_from_header(header)

            if type_found:
                ob.type = type_or_var
            else:
                for variable in ob.variables:
                    if variable.name.lower() == type_or_var.lower():
                        ob.type = variable.type
                        break
        self.objects.append(ob)

    def find_routines(self):
        # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
        # dimensions and optionality of the arguments.

        # Look into each line
        for _index, line in enumerate(self.lines):
            # Look if routine starts with subroutine
            if re.search("subroutine", line, re.IGNORECASE):
                if line.lower().count("wclock_"):
                    pass
                if line.lower().strip().find("end") == 0:
                    continue
                routine_start = _index
                # Look for end subroutine statements
                routine_end = self.find_next_routine_end(_index)

                # Find name and dummy arguments
                routine_name, dummy_arguments = self.get_dummy_arguments(_index)
                # Find declarations
                declarations = self.find_routine_declarations(routine_start, routine_end)
                header = self.lines[routine_start].lower()
                # Parse everything and store
                self.find_procedure_variables(routine_name.lower(), dummy_arguments, declarations, header)

    def find_types(self):
        # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
        # dimensions and optionality of the arguments.

        # Look into each line
        in_type = False
        types = {}
        type_name = None
        variable_counter = None
        for _index, line in enumerate(self.lines):
            # Look if routine starts with subroutine
            if not in_type:
                pattern = "^ *type[^()]*[ \t](\w+) *$"
                m = re.search(pattern, line, re.IGNORECASE)
                if m:
                    in_type = True
                    variable_counter = 0
                    type_name = m.groups(0)[0]
                    types[type_name] = {}
                    ob = DerivedType(type_name, self.module)
                    self.objects.append(ob)
            else:
                pattern = "^ *END TYPE"
                m = re.search(pattern, line, re.IGNORECASE)
                if m:
                    in_type = False
                else:
                    pattern = "(.*)::(.*)"
                    m = re.search(pattern, line)
                    if m:
                        attributes, variables = m.groups()
                        var_type, dimension, intent, optional, allocatable, is_parameter = parse_attributes(attributes)
                        variables = split_declaration_arguments(variables)
                        for var in variables:
                            types[type_name][var] = var_type
                            v = Variable(ob, var, var_type, dimension)
                            v.position = variable_counter
                            v.is_dummy_argument = True
                            v.allocatable = allocatable
                            v.is_parameter = is_parameter
                            ob.add_variable(v)
                            variable_counter += 1
        return types

    def find_module_variables(self):
        main_routine = ModuleMain(self.filename)
        self.objects.append(main_routine)
        pattern = "(.*)::(.*)"
        for line in self.lines:
            if line.lower().count("contains"):
                break
            if line.lower().count("end program"):
                break
            match = re.search(pattern, line)
            if match:
                attributes, variables = match.groups()
                var_type, dimension, intent, optional, allocatable, is_parameter = parse_attributes(attributes)
                variables = split_declaration_arguments(variables)
                if var_type:
                    for variable in variables:
                        v = Variable(main_routine, variable, var_type, dimension)
                        v.allocatable = allocatable
                        v.is_parameter = is_parameter
                        main_routine.add_variable(v)

    def is_start_of_function_definition(self, line_number):
        line = self.lines[line_number]
        match = re.search("(^ *[a-z\(\)0-9=]* *function.*)", line, re.I)
        if match:
            if match.group().lower().count("end function"):
                return False
            return True
        return False

    def find_functions(self):
        # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
        # dimensions and optionality of the arguments.

        # Look into each line
        for _index, line in enumerate(self.lines):
            if self.is_start_of_function_definition(_index):
                function_start = _index
                # Look for end subroutine statements
                function_end = self.find_next_function_end(_index)

                # Find name and dummy arguments
                function_name, dummy_arguments = self.get_dummy_arguments(_index)
                # Find declarations
                declarations = self.find_routine_declarations(function_start, function_end)
                # Parse everything and store
                header = self.lines[function_start].lower()
                self.find_procedure_variables(function_name.lower(), dummy_arguments, declarations, header)

    def find_implicit_functions(self, vault):
        try:
            routine = self.module.main()
        except AttributeError:
            return None
        # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
        # dimensions and optionality of the arguments.
        pattern = "^ *(\w+) *\([\w, ]*\) *= *(.*)"
        # Look into each line
        for _index, line in enumerate(self.lines):
            m = re.search(pattern, line)
            if m:

                call = close_brackets(line[m.start(1):])
                call_arguments = find_call_arguments(call)
                is_function = False
                for argument in call_arguments:
                    argument_type = type_of_contents(argument, routine=routine, vault=vault)
                    if argument_type != "integer":
                        is_function = True
                        break

                if is_function:
                    function_name = find_called_function(call)
                    new_function = Function(function_name, module=self.module)
                    new_function.type = type_of_contents(function_name, routine=routine, vault=vault)
                    for argument_index, argument in enumerate(call_arguments):
                        var = routine.get_variable_by_name(argument)
                        var.is_dummy_argument = True
                        var.position = argument_index
                        var.intent = "in"
                        new_function.add_variable(var)
                    self.objects.append(new_function)

            else:
                routine = in_which_block(line, _routine=routine, vault=vault)

    def find_interfaces(self, object_procedures):
        # Definition: Looks into the sourcefile for interface declarations, storing the name.
        for _index, _line in enumerate(self.lines):
            if _line.strip().lower().find("interface") == 0:
                interface_start = _index
                interface_end = self.find_next_interface_end(_index)
                interface_name = _line.split()[1].strip()
                routines = []
                line_routines = None
                for interface_line in self.lines[interface_start:interface_end + 1]:
                    if interface_line.strip().lower().find("module procedure") == 0:
                        line_routines = interface_line.lower().replace("module procedure", "")
                        line_routines = [_r.strip() for _r in line_routines.split(",")]
                        routines += line_routines
                    elif interface_line.strip().lower().strip().find("subroutine") == 0:
                        routine_name = interface_line[:interface_line.find("(")].split()[1]
                        routines += [routine_name]

                interface_procedures = [object_procedures[_routine] for _routine in routines]
                ob = Interface(interface_name)
                for procedure in interface_procedures:
                    ob.add_procedure(procedure)
                if not ob.procedures:
                    raise AssertionError("No routine found for interface %s" % interface_name)
                self.objects.append(ob)

    def find_used_modules(self, vault):
        current_block = self.module.main()
        for line in self.lines:
            current_block = in_which_block(line, current_block, vault=vault)
            line = line.strip().lower()
            if line.find("use ") == 0:
                module_name = line.split()[1]
                if module_name.count(","):
                    module_name = module_name.replace(",", "")
                for module in vault.modules:
                    if module.name == module_name:
                        self.module.uses.append(module)

                # Look for
                pattern = "(\w+) *=> *(\w+)"
                matches = re.findall(pattern, line)
                if matches:
                    for match in matches:
                        pointer, target = match
                        target_var = self.module.uses[-1].main().get_variable_by_name(target)
                        new_var = Variable(current_block, pointer, target_var.type, target_var.dimension)
                        current_block.add_variable(new_var)


class ImplementRPE(ReadSourceFile):
    def __init__(self, _filename, vault):
        super(ImplementRPE, self).__init__(_filename)

        # Dictionaries and lists about source info
        self.source_info = vault
        if self.source_info:
            self.function_names = [fun.name for fun in self.source_info.functions]
            module_name = _filename.replace(".f90", "")
            if module_name.count("/"):
                module_name = module_name.split("/")[-1]
            self.module = self.source_info[module_name]
        else:
            self.function_names = []
            self.module = None

    def pre_process(self, replace_wrk_alloc=True, replace_reals=True):
        # Replace wrk_alloc
        # TODO: This is only useful with NEMO and probably for old versions
        if replace_wrk_alloc:
            self.replace_wrk_alloc()
        if replace_reals:
            # Replace real declarations with type(rpe_var)
            self.replace_real_declaration()
            # Workaround for parameters
            self.replace_parameter_declaration()

    def post_process(self):
        # If we don't have source info there's nothing to process
        if not self.source_info:
            return None

        # Process intrinsic calls
        self.fix_intrinsic_calls()
        # Process routine calls
        self.fix_routines()
        # Process function calls
        self.fix_functions()

        self.fix_type_conversion()

        # Process read and write calls
        self.fix_read_and_write_calls()
        # Fix pointer assignations
        self.fix_pointer_assignations()
        # Add use rp_emulator statement
        self.add_use_rpe_statement()

    def fix_pointer_assignations(self):
        current_block = self.module.main()
        pattern = "^[^:'\n\w]*([\w%\(\):,]+) *=> *([\w%\(\):, -]*)"

        for line_index, line in enumerate(self.lines):
            if not re.search(pattern, line):
                current_block = in_which_block(line, current_block, vault=self.source_info)
            else:
                m = re.search(pattern, line)
                pointer, target = m.groups(0)
                pointer_type = type_of_contents(pointer, routine=current_block, vault=self.source_info)
                target_type = type_of_contents(target, routine=current_block, vault=self.source_info)
                if pointer_type == "real" and target_type == "rpe_var":
                    self.lines[line_index] = line.replace(target, "%s%%val" % target)

    def replace_wrk_alloc(self):
        self.text = self.text.replace("USE wrk_nemo", "")
        self.text = replace_wrkalloc(self.text)
        self.lines = self.text.split("\n")

    def replace_real_declaration(self):
        # Look at each line for REAL variable declarations and replace it with type(rpe_var)
        for _index, line in enumerate(self.lines):
            _match = re.search('[^\!]*real.*::', line, flags=re.I)
            if _match:
                _match = re.search('[^\!]*(real *\([\w= ]*\)).*::', line, flags=re.I)
                if _match:
                    real_type = _match.groups(0)[0]
                    line = line.replace(real_type, "type(rpe_var)")
                    self.lines[_index] = line
                else:
                    _match = re.search('[^\!]*(real).*::', line, flags=re.I)
                    real_type = _match.groups(0)[0]
                    line = line.replace(real_type, "type(rpe_var)")
                    self.lines[_index] = line
        self.rebuild_text()

    def replace_parameter_declaration(self):
        # If there's an assignation whe have to find a workaround
        for _index, line in enumerate(self.lines):
            if re.search('^[^\!]*type\(rpe_var\).*::.*=', line, flags=re.I):
                arguments = line.split("::")[1]

                type_of_exception = find_kind_of_parameter_exception(arguments)

                if type_of_exception == "regular":
                    piece_to_replace = arguments.split("=")[1].strip()

                    replacement = " rpe_var( 52, %s ) " % piece_to_replace
                    self.lines[_index] = line.replace(piece_to_replace, replacement)
                elif type_of_exception == "multiple_variables":
                    for argument in arguments.split(","):
                        piece_to_replace = argument.split("=")[1].strip()
                        replacement = " rpe_var(  52, %s ) " % piece_to_replace
                        line = line.replace(piece_to_replace, replacement)
                    self.lines[_index] = line
                elif type_of_exception == "array_initialization":
                    piece_to_replace = arguments
                    replacement = add_rpe_var_to_arguments(arguments)
                    self.lines[_index] = line.replace(piece_to_replace, replacement)
                elif type_of_exception == "already_fixed":
                    continue
        self.rebuild_text()

    def add_use_rpe_statement(self):
        # Adding module rp_emulator (in theory adding this to par_kind.f90 should be enough,
        # but theory fails and compiler complains so I'll add it everywhere
        self.lines.insert(1, "USE rp_emulator")
        self.rebuild_text()

    def fix_read_and_write_calls(self):
        self.fix_read_calls()
        self.fix_write_calls()

    def fix_read_calls(self):
        current_block = self.module.main()
        pattern = "[^\w](read *\(.*)"
        for line_index, line in enumerate(self.lines):
            matches = re.search(pattern, line, re.IGNORECASE)
            if matches:

                to_remove = close_brackets(matches.groups(1)[0])
                read_arguments = line[matches.start(1):].replace(to_remove, "")

                arguments = split_read_arguments(read_arguments)

                for argument_index, argument in enumerate(arguments):
                    argument_type = type_of_contents(argument, routine=current_block, vault=self.source_info)
                    if argument_type == "rpe_var":
                        arguments[argument_index] = "%s%%val" % argument
                replacement = ",".join(arguments)
                self.lines[line_index] = line.replace(read_arguments, replacement)
            else:
                current_block = in_which_block(line, current_block, vault=self.source_info)

    def fix_write_calls(self):
        current_block = self.module.main()
        pattern = "^[^\!]* +write\([^()]*\) (.*)"

        for line_index, line in enumerate(self.lines):
            if line_index == 133:
                pass
            if re.search("WRITE *\(", line):
                wc = close_brackets(line[line.find("WRITE"):])

                line = line.replace(wc, "WRITE()")
            matches = re.search(pattern, line, re.IGNORECASE)
            if matches:
                to_replace = matches.groups(0)[0]
                arguments = split_write_arguments(line)
                if not arguments:
                    continue
                for argument_index, argument in enumerate(arguments):
                    argument_type = type_of_contents(argument, routine=current_block, vault=self.source_info)
                    if argument_type == "rpe_var":
                        arguments[argument_index] = "real( %s )" % argument
                if arguments[0] == "(":
                    arguments = arguments[1:-1]
                    replacement = "( %s )" % ",".join(arguments)
                else:
                    replacement = ",".join(arguments)
                self.lines[line_index] = self.lines[line_index].replace(to_replace, replacement)
            else:
                current_block = in_which_block(line, current_block, vault=self.source_info)

    def fix_routines(self):
        current_block = self.module.main()
        for _index, _line in enumerate(self.lines):
            # Find routine calls in line
            routine_calls = calls_to_routines(_line)
            if routine_calls is None:
                # If not, look if the code is entering or exiting a routine block
                current_block = in_which_block(_line, current_block, vault=self.source_info)
            else:
                temporal_line = _line
                for call in routine_calls:
                    is_external = False
                    called_routine = find_called_routine(call)
                    called_arguments = find_call_arguments(call)
                    if not called_arguments:
                        continue
                    procedure = self.source_info.get_procedure(called_routine)
                    if isinstance(procedure, Routine):
                        pass
                    elif isinstance(procedure, Interface):
                        procedure = match_arguments_and_interface(called_arguments, interface=procedure,
                                                                  current_block=current_block, vault=self.source_info)
                    else:
                        warnings.warn("Routine not found in vault database %s" % called_routine)
                        is_external = True
                        # continue

                    temporal_call = call
                    if not is_external:
                        for argument_index, argument in enumerate(called_arguments):
                            if is_keyword_argument(argument):
                                keyword = argument.split("=")[0].strip()
                                argument_type = type_of_contents(argument, routine=current_block,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_name(keyword)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found: %s" % keyword)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type == "real" and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                                elif argument_type == "rpe_var" and dummy_argument.type == "real":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                            else:
                                argument_type = type_of_contents(argument, routine=current_block,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_position(argument_index)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found in position: %i" % argument_index)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type == "real" and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError(
                                            "Types do not coincide but there's no easy workaround here:\n %s" % call)
                                elif argument_type == "rpe_var" and dummy_argument.type == "real":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "inout":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                        temporal_line = temporal_line.replace(call, temporal_call)
                    else:
                        for argument_index, argument in enumerate(called_arguments):
                            # TODO: Probably should find a better solution, the code should not identify an array as a
                            # Todo: ... function even if the name coincides.
                            # if argument == ":":
                            #     continue
                            argument_type = type_of_contents(argument, routine=current_block, vault=self.source_info)
                            if argument_type == "rpe_var":
                                # TODO: Probably I should check first if the argument is a single variable and can be used as an output argument or not
                                if re.match("^\w+$", argument):
                                    temporal_call = add_val_to_argument(temporal_call, argument_index)
                                elif re.match("^\w+ *\([^\(\)]*\)$", argument):
                                    temporal_call = add_val_to_argument(temporal_call, argument_index)
                                elif not has_operations(argument):
                                    temporal_call = add_val_to_argument(temporal_call, argument_index)
                                else:
                                    temporal_call = add_real_to_argument(temporal_call, argument_index)
                        temporal_line = temporal_line.replace(call, temporal_call)
                self.lines[_index] = temporal_line
        self.rebuild_text()

    def fix_functions(self):
        current_block = self.module.main()
        for line_index, line in enumerate(self.lines):
            function_calls = calls_to_functions(line, self.source_info)
            if not function_calls:
                # If not, look if the code is entering or exiting a routine block
                current_block = in_which_block(line, current_block, vault=self.source_info)
            else:
                # Fix function calls (line_index
                temporal_line = line
                for call in function_calls:
                    is_external = False
                    called_function = find_called_function(call)
                    if called_function.lower() == current_block.name.lower():
                        continue
                    called_arguments = find_call_arguments(call)
                    if not called_arguments:
                        continue

                    procedure = self.source_info.get_procedure(called_function)
                    if isinstance(procedure, Function):
                        pass
                    elif isinstance(procedure, Interface):
                        procedure = match_arguments_and_interface(called_arguments, interface=procedure,
                                                                  current_block=current_block, vault=self.source_info)
                    else:
                        warnings.warn(
                            "Function not found in vault database %s\nWill be treated as external." % called_function)
                        is_external = True
                        # continue

                    if not is_external:
                        temporal_call = call
                        for argument_index, argument in enumerate(called_arguments):
                            # TODO: Probably should find a better solution, the code should not identify an array as a
                            # Todo: ... function even if the name coincides. sbcice_lim and lib_fortran fail without it
                            if argument == ":":
                                continue
                            if is_keyword_argument(argument):
                                keyword = argument.split("=")[0].strip()
                                argument_type = type_of_contents(argument, routine=current_block,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_name(keyword)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found: %s" % keyword)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type == "real" and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                                elif argument_type == "rpe_var" and dummy_argument.type == "real":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                            else:
                                argument_type = type_of_contents(argument, routine=current_block,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_position(argument_index)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found in position: %i" % argument_index)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type == "real" and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                                elif argument_type == "rpe_var" and dummy_argument.type == "real":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                        temporal_line = temporal_line.replace(call, temporal_call)
                    else:
                        temporal_call = call
                        for argument_index, argument in enumerate(called_arguments):
                            # TODO: Probably should find a better solution, the code should not identify an array as a
                            # Todo: ... function even if the name coincides.
                            # if argument == ":":
                            #     continue
                            argument_type = type_of_contents(argument, routine=current_block, vault=self.source_info)
                            if argument_type == "rpe_var":
                                # TODO: Probably I should check first if the argument is a single variable and can be used as an output argument or not
                                temporal_call = add_val_to_argument(temporal_call, argument_index)
                        temporal_line = temporal_line.replace(call, temporal_call)
                self.lines[line_index] = temporal_line
        self.rebuild_text()

    def fix_type_conversion(self):
        current_block = self.module.main()
        for line_index, line in enumerate(self.lines):
            function_calls = calls_to_procedure(line, procedure_type=DerivedType, vault=self.source_info)
            if not function_calls:
                # If not, look if the code is entering or exiting a routine block
                current_block = in_which_block(line, current_block, vault=self.source_info)
            else:
                # Fix function calls (line_index
                temporal_line = line
                for call in function_calls:
                    called_function = find_called_function(call)
                    called_arguments = find_call_arguments(call)
                    if not called_arguments:
                        continue

                    procedure = self.source_info.get_procedure(called_function)
                    if isinstance(procedure, DerivedType):
                        pass
                    else:
                        raise TypeError("The %s procedure does not have the DerivedType type" % procedure.name)

                    temporal_call = call
                    for argument_index, argument in enumerate(called_arguments):
                        # TODO: Probably should find a better solution, the code should not identify an array as a
                        # Todo: ... function even if the name coincides. sbcice_lim and lib_fortran fail without it
                        if argument == ":":
                            continue
                        argument_type = type_of_contents(argument, routine=current_block,
                                                         vault=self.source_info)
                        dummy_argument = procedure.get_variable_by_position(argument_index)
                        if not dummy_argument:
                            raise AssertionError("Variable not found in position: %i" % argument_index)
                        if argument_type == dummy_argument.type:
                            continue
                        elif argument_type == "real" and dummy_argument.type == "rpe_var":
                            temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                        else:
                            raise AssertionError("Types do not coincide and the case is not considered %s %s" % (
                                argument_type, dummy_argument.type))
                    temporal_line = temporal_line.replace(call, temporal_call)
                self.lines[line_index] = temporal_line
        self.rebuild_text()

    def fix_intrinsic_calls(self):
        # todo: this intrinsic calls only need to be corrected if there's a hardcoded real
        list_of_intrinsics = ["max", "min", "max_loc", "min_loc", "cmplx", "sum"]
        for intrinsic in list_of_intrinsics:
            for _index, _line in enumerate(self.lines):
                # This may seem redundant but as we are modifying the lines multiple times
                # we need to be sure that the line is up to date
                _line = self.lines[_index]
                fun = "[ ,(+-](%s *)\(" % intrinsic

                matches = re.finditer(fun, _line, flags=re.I)
                matches = [m for m in matches]

                for _i in range(matches.__len__()):
                    matches = re.finditer(fun, _line, flags=re.I)
                    matches = [m for m in matches]
                    m = matches[_i]

                    start = m.start(0)
                    piece_to_replace = _line[start:]

                    # Bug fix: first char should'nt be a bracket
                    start = piece_to_replace.lower().find(intrinsic)

                    piece_to_replace = piece_to_replace[start:]

                    piece_to_replace = close_brackets(piece_to_replace)
                    # if there's no point (dirty way to look for a hardcoded real) we can skip this transformation
                    if not piece_to_replace.count("."):
                        continue

                    number_of_arguments, intervals = find_argument_intervals(piece_to_replace)
                    replacement = piece_to_replace
                    if intrinsic == "cmplx":
                        number_of_arguments = 2
                    for __i in range(number_of_arguments):
                        replacement = add_real_to_argument(replacement, __i)
                    if intrinsic == "sum":
                        pass
                    _line = _line.replace(piece_to_replace, replacement)

                self.lines[_index] = _line


class RemoteFolder:
    def __init__(self, input_sources, output_sources, _remote_folder):
        self.machine = "nord3"
        self.input_sources = input_sources
        self.output_sources = output_sources
        self.remote_folder = _remote_folder

    def pull_sources_from_remote(self, filenames=None):
        if not filenames:
            filenames = "*"
        sync_command = "rsync -rvc %s:%s/%s %s" % (self.machine, self.remote_folder, filenames, self.input_sources)
        system(sync_command)

    def push_sources_to_remote(self):
        sync_command = "rsync -rvc %s/* %s:%s" % (self.output_sources, self.machine, self.remote_folder)
        system(sync_command)

    def compile_nemo(self, configuration):
        compilation_script_name = "compilation_script.sh"

        if self.machine == "nord3":
            environment_file = "gcc-openmpi.env"
            arch_file = "nord3-gcc4.9.1-openmpi1.8.1-debug"
        elif self.machine == "mn4":
            environment_file = "mn4.env"
            arch_file = "mn4-debug"
        else:
            raise AssertionError("Unknown machine %s" % self.machine)

        compilation_script = '''
#!/usr/bin/env bash
# Go to folder
cd /gpfs/projects/bsc32/bsc32870/NEMO/latest/CONFIG

# Configuration to compile with RPE
# CONFIG=O2L3-nord3-rpe-scripts-2.0
CONFIG=%s
ENV_FILE=%s
ARCH_FILE=%s
# Copy rpe .mod and .o
./put_rpe.cmd ${CONFIG}


# Set environment
source ${ENV_FILE}

#Compile
./makenemo -n ${CONFIG} -m ${ARCH_FILE} -j 16
''' % (configuration, environment_file, arch_file)

        with open(compilation_script_name, "w") as script_file:
            script_file.write(compilation_script)
        system("ssh %s 'bash -s' < %s " % (self.machine, compilation_script_name))


def parse_sources(path_to_model_sources):
    # Find the nemo source-files in sources
    # (Using the sources that already modified replacing real declarations by type(rpe_var) )
    nemo_sources = [f for f in listdir(path_to_model_sources) if
                    isfile(join(path_to_model_sources, f)) and f.count(".f90")]

    # Initialize two empty dictionaries that will contain all the routines and all the functions
    object_oriented_dictionary = {"procedures": {}, "interfaces": {}, "variables": {}, "modules": {},
                                  "derivedtypes": {}}

    # Initialize an empty dictionary that will contain all the source-files.
    sourcefiles = {}
    vault = Vault()

    # Going over all the nemo source-files loading the files,
    # parsing the routines and storing both things in dictionaries.
    for index, filename in enumerate(nemo_sources):
        module_name = filename.replace(".f90", "")
        sourcefiles[module_name] = ObtainSourcefileInfo(join(path_to_model_sources, filename))
        module = Module(module_name)
        f = sourcefiles[module_name]
        f.module = module
        f.find_routines()
        f.find_functions()
        f.find_module_variables()
        for procedure in f.objects:
            object_oriented_dictionary["procedures"][procedure.name] = procedure
            module.add_procedure(procedure)
        vault.add_module(module)

    # Going over all the nemo source-files parsing the interfaces and storing them in dictionary.
    for index, filename in enumerate(nemo_sources):
        module_name = filename.replace(".f90", "")
        f = sourcefiles[module_name]
        f.find_interfaces(object_oriented_dictionary["procedures"])
        # Getting and saving modules used
        f.module = vault[module_name]
        f.find_used_modules(vault)
        for interface in f.objects:
            if isinstance(interface, Interface):
                object_oriented_dictionary["interfaces"][interface.name] = interface
                vault[module_name].procedures.append(interface)

    for filename in sourcefiles:
        module_name = filename.replace(".f90", "")
        f = sourcefiles[module_name]
        f.find_types()
        for derived_type in f.objects:
            if isinstance(derived_type, DerivedType):
                object_oriented_dictionary["derivedtypes"][derived_type.name] = derived_type
                vault[module_name].procedures.append(derived_type)

    vault.process_variables()

    # Find implicit function declarations
    for filename in sourcefiles:
        module_name = filename.replace(".f90", "")
        f = sourcefiles[module_name]
        f.find_implicit_functions(vault=vault)
        for procedure in f.objects:
            if procedure not in vault[module_name].procedures:
                vault[module_name].procedures.append(procedure)

    vault.process_procedures()
    # Search which modules does each module has access to
    vault.expand_used_modules()
    # Search which modules have access to this module
    vault.expand_used_by()

    save = True

    if save:
        # This workaround is needed to allow pickle to write the file, don't know yet about its risks
        import sys
        sys.setrecursionlimit(2000)
        pickle.dump(vault, open("vault.pkl", "wb"))


def process_individual_file(path_to_input_sources, path_to_output_sources, vault, filename, preprocess=True):
    # This function loads a file and applies modifications to the file.
    # TODO: Find an automatic way of selecting routines to blacklist based on calls to external routines
    # Files to ignore

    f = ImplementRPE(join(path_to_input_sources, filename), vault)

    if preprocess:
        f.pre_process()
    else:
        f.pre_process(replace_reals=False)
    f.post_process()

    # Several Workarounds for the NEMO case --------------------------------
    # TODO: Make all these workarounds dispensable
    WORKAROUNDS = True
    if WORKAROUNDS:
        if filename == "limhdf.f90":
            # TODO: Scan for dependencies and see if any variable is used as a target(?)
            f.text = f.text.replace("type(rpe_var), ALLOCATABLE , DIMENSION(:,:,:)  :: zrlx",
                                    "type(rpe_var), ALLOCATABLE , DIMENSION(:,:,:), TARGET ::  zrlx")
        if filename == "fldread.f90":
            # TODO: Probably most of the exceptions can be removed, the last should be solved in the source code.
            pairs = [["zlsm2d = (SUM ( zlsm3d, 3 , ll_msknan3d ) )/(MAX(1,(COUNT( ll_msknan3d , 3 ))   ))",
                      "zlsm2d = (SUM ( zlsm3d%val, 3 , ll_msknan3d ) )/(MAX(1,(COUNT( ll_msknan3d , 3 ))   ))"],
                     ["IF( ASSOCIATED(ref_wgts(kw)%data_wgt) )  WRITE(numout,*)",
                      "IF( ALLOCATED(ref_wgts(kw)%data_wgt) )  WRITE(numout,*)"],
                     ["INTEGER, ALLOCATABLE, DIMENSION(:)  ::  ptr",
                      "INTEGER, POINTER, DIMENSION(:)  ::  ptr"]]
            for pair in pairs:
                f.text = f.text.replace(pair[0], pair[1])
        if filename == "domzgr.f90":
            # TODO: Bug in the NEMO sourcecode
            pair = ["CALL prihre(  hbatt(1,1)%val",
                    "CALL prihre(  hbatt(:,:)%val"]
            f.text = f.text.replace(pair[0], pair[1])
        if filename == "sbccpl.f90":
            # TODO: Don't know if it can be removed. It seems that it can't
            pairs = [["srcv( (/jpr_otx1,jpr_oty1/) )%nsgn = -1.",
                      "srcv( (/jpr_otx1,jpr_oty1/) )%nsgn = rpe(-1.)"]]
            for pair in pairs:
                f.text = f.text.replace(pair[0], pair[1])
        if filename == "sbcblk_mfs.f90":
            # TODO: Should look closer at why this workaround is needed
            pairs = [["type(rpe_var) :: onsea,par1,par2",
                      "real(wp) :: onsea,par1,par2"],
                     ["type(rpe_var), DIMENSION(12) :: alpham",
                      "real(wp), DIMENSION(12) :: alpham"],
                     ["IF(coszen .LE. 5.035D-04) THEN",
                      "IF(coszen .LE. 5.035E-04) THEN"]
                     ]
            for pair in pairs:
                f.text = f.text.replace(pair[0], pair[1])
        if filename == "sbcblk_clio.f90":
            # TODO: Don't know if it can be removed. These variables use DATA statements with real values and the workaround is not straight-forward
            pairs = [["type(rpe_var), DIMENSION(20)  :: tauco",
                      "real(wp), DIMENSION(20)  :: tauco"],
                     ["type(rpe_var), DIMENSION(19)  ::  budyko",
                      "real(wp), DIMENSION(19)  ::  budyko"]]
            for pair in pairs:
                f.text = f.text.replace(pair[0], pair[1])
        if filename == "geo2ocean.f90":
            # TODO: Don't know if it can be removed
            pairs = [["type(rpe_var), PARAMETER :: rad =  rpe_var( 52, rpi / 180.e0 )",
                      "type(rpe_var), PARAMETER :: rad =  rpe_var( 52, rpi%val / 180.e0 )"]]
            for pair in pairs:
                f.text = f.text.replace(pair[0], pair[1])

    # Save the new file in the Modified sources folder
    f.save_modified_file(join(path_to_output_sources, filename))


def process_sources(path_to_input_sources, path_to_output_sources, vault, preprocess_blacklist=None):
    # Find files in sources
    all_sources = [f for f in listdir(path_to_input_sources) if
                   isfile(join(path_to_input_sources, f)) and f.count(".f90")]

    # filter_name = 'sbcapr.f90'
    # all_sources = [f for f in all_sources if f == filter_name]

    # Process Sources!
    for index, filename in enumerate(all_sources):
        print("Processing %s" % filename)
        if preprocess_blacklist:
            if filename in preprocess_blacklist:
                preprocess = False
            else:
                preprocess = True
        else:
            preprocess = True
        process_individual_file(path_to_input_sources, path_to_output_sources, vault, filename, preprocess)
        print("%s done!" % filename)


class Counter:
    def __init__(self):
        self.count = 0

    def up(self, how_much=1):
        self.count += how_much

    def down(self, how_much=1):
        self.count -= how_much


def find_var_elsewhere(variable, file_objects):
    pattern = "[^\w]%s[^\w]"

    files = ["%s.f90" % module.name for module in variable.routine.module.used_by]
    files.append("%s.f90" % variable.routine.module.name)
    files_to_search = []
    for f in file_objects:
        if f in files:
            files_to_search.append(file_objects[f])
        if len(files_to_search) == files:
            break

    return [f for f in files_to_search if re.search(pattern % variable.name, f.text)]


def find_file_and_line_of_allocation(variable, file_objects):
    if is_allocated_inside_declaration_scope(variable, file_objects):
        f = file_objects["%s.f90" % variable.routine.module.name]
        for line_index, line in enumerate(f.lines):
            if re.search("allocate *\(.*[^\w]%s[^\w]" % variable.name, line, re.IGNORECASE):
                return f, line_index
        raise ValueError("Allocation of variable %s not found" % variable.name)
    else:
        files_where_var_appears = find_var_elsewhere(variable, file_objects)
        for f in files_where_var_appears:
            for line_index, line in enumerate(f.lines):
                if re.search("allocate *\(.*[^\w]%s[^\w]" % variable.name, line, re.IGNORECASE):
                    return f, line_index
        warnings.warn("Allocation of variable %s not found" % variable.name)
        return None, None


def find_file_and_line_of_namelist_reading(variable, file_objects):
    routine = variable.routine
    module = routine.module
    sourcefile = file_objects["%s.f90" % module.name]

    pattern = "namelist.*[^\w]%s[^\w]" % variable.name
    namelist_name = None
    for line in sourcefile.lines:
        if re.search(pattern, line, re.IGNORECASE):
            pattern2 = "namelist/(.*)/.*[^\w]%s[^\w]" % variable.name
            match = re.search(pattern2, line, re.IGNORECASE)
            namelist_name = match.group(1)
            break
    if namelist_name:
        for index, line in enumerate(sourcefile.lines):
            pattern3 = "read *\(.*[^\w]%s[^\w]" % namelist_name
            if re.search(pattern3, line, re.IGNORECASE):
                return sourcefile, index

    other_sources = find_var_elsewhere(variable, file_objects)
    for sourcefile in other_sources:
        pattern = "namelist.*[^\w]%s[^\w]" % variable.name
        namelist_name = None
        for line in sourcefile.lines:
            if re.search(pattern, line, re.IGNORECASE):
                pattern2 = "namelist/(.*)/.*[^\w]%s[^\w]" % variable.name
                match = re.search(pattern2, line, re.IGNORECASE)
                namelist_name = match.group(1)
                break
        if namelist_name:
            for index, line in enumerate(sourcefile.lines):
                pattern3 = "read *\(.*[^\w]%s[^\w]" % namelist_name
                if re.search(pattern3, line, re.IGNORECASE):
                    return sourcefile, index

    return None, None


def find_first_use(variable, file_objects):
    routine = variable.routine
    module = routine.module

    sourcefile = file_objects["%s.f90" % module.name]

    pattern = "[^\w]%s[^\w]" % variable.name
    searching = False
    loop_level = 0
    loop_start = None
    for line_index, line in enumerate(sourcefile.lines):
        if not searching:
            if line.lower().count("subroutine"):
                called_routine = re.search("subroutine *(\w*)", line, re.IGNORECASE).group(1)
                if called_routine.lower() == routine.name.lower():
                    searching = True
        else:
            if re.search(pattern, line, re.IGNORECASE):
                if line.count("::"):
                    pass
                elif line.lower().count("subroutine"):
                    pass
                elif line.lower().find("use ") == 0:
                    pass
                else:
                    if loop_level > 0:
                        return sourcefile, loop_start
                    else:
                        return sourcefile, line_index
            elif line.lower().count("end subroutine"):
                break
            elif re.search("^ *do *[\w]* *=", line, re.IGNORECASE):
                if loop_level == 0:
                    loop_start = line_index
                loop_level += 1
            elif re.search("^ *end +do", line, re.IGNORECASE):
                loop_level -= 1
    return None, None


def insert_variable_precision_specification(variable, sourcefile, insertion_line):
    new_line = "    %s%%sbits = emulator_variable_precisions(%i)" % (variable.name, variable.id)
    sourcefile.lines.insert(insertion_line, new_line)
    sourcefile.rebuild_text()


def insert_variable_truncation(variable, sourcefile, insertion_line):
    new_line = "    CALL apply_truncation(%s)" % variable.name
    sourcefile.lines.insert(insertion_line, new_line)
    sourcefile.rebuild_text()


def insert_variable_precision_specification_to_allocatable(variable, file_objects, counter_object):
    sourcefile, allocation_line = find_file_and_line_of_allocation(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        counter_object.up()
        insert_variable_precision_specification(variable, sourcefile, allocation_line + 1)


def insert_variable_precision_specification_to_namelist_parameter(variable, file_objects, counter_object):
    sourcefile, reading_line = find_file_and_line_of_namelist_reading(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        counter_object.up()
        insert_variable_precision_specification(variable, sourcefile, reading_line + 1)
    else:
        warnings.warn("Variable %s not found" % variable.name)


def insert_variable_precision_to_routine_variable(variable, file_objects, counter_object):
    sourcefile, first_use_line = find_first_use(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        counter_object.up()
        insert_variable_precision_specification(variable, sourcefile, first_use_line)
        insert_variable_truncation(variable, sourcefile, first_use_line + 1)
    else:
        pass
        warnings.warn("Variable %s not found" % variable.name)


def insert_variable_precision_after_routine_call(path_to_sources, vault=None, counter=None):
    source_files = load_source_files(path_to_sources)
    for name, source_file in source_files.items():
        # print name
        add_after_call_routines(source_file, vault, counter)
        source_file.rebuild_text()
        source_file.save_modified_file(join(path_to_sources, source_file.filename))


def add_after_call_routines(source_file, vault, counter):
    current_block = None

    for _index, _line in enumerate(source_file.lines):
        # Find routine calls in line
        routine_calls = calls_to_routines(_line)
        if routine_calls is None:
            # If not, look if the code is entering or exiting a routine block
            current_block = in_which_block(_line, current_block, vault=vault)
        else:
            for call in routine_calls:
                is_external = False
                called_routine = find_called_routine(call)
                if called_routine == "apply_truncation":
                    continue
                called_arguments = find_call_arguments(call)
                if not called_arguments:
                    continue
                procedure = vault.get_procedure(called_routine)
                if isinstance(procedure, Routine):
                    pass
                elif isinstance(procedure, Interface):
                    procedure = match_arguments_and_interface(called_arguments, interface=procedure,
                                                              current_block=current_block, vault=vault)
                else:
                    warnings.warn("Routine not found in vault database %s" % called_routine)
                    is_external = True
                    # continue

                if not is_external:
                    for argument_index, argument in enumerate(called_arguments):
                        if is_keyword_argument(argument):
                            keyword = argument.split("=")[0].strip()
                            dummy_argument = procedure.get_variable_by_name(keyword)
                            if not dummy_argument:
                                raise AssertionError("Variable not found: %s" % keyword)
                            if dummy_argument.type != "rpe_var":
                                if dummy_argument.type == "real" and argument.count("%val"):
                                    var = retrive_variable(argument, routine=current_block, vault=vault)
                                    insert_in_line(var, source_file, counter, _index)

                                continue
                            if dummy_argument.intent == "out" or dummy_argument.intent == "inout":
                                var = retrive_variable(argument, routine=current_block, vault=vault)
                                insert_in_line(var, source_file, counter, _index)

                        else:
                            dummy_argument = procedure.get_variable_by_position(argument_index)
                            if not dummy_argument:
                                raise AssertionError("Variable not found in position: %i" % argument_index)
                            if dummy_argument.type != "rpe_var":
                                if dummy_argument.type == "real" and argument.count("%val"):
                                    var = retrive_variable(argument, routine=current_block, vault=vault)
                                    insert_in_line(var, source_file, counter, _index)
                                continue
                            if dummy_argument.intent == "out" or dummy_argument.intent == "inout":
                                var = retrive_variable(argument, routine=current_block, vault=vault)
                                insert_in_line(var, source_file, counter, _index)

                else:
                    for argument_index, argument in enumerate(called_arguments):
                        # TODO: Probably should find a better solution, the code should not identify an array as a
                        # Todo: ... function even if the name coincides.
                        # if argument == ":":
                        #     continue
                        argument_type = type_of_contents(argument, routine=current_block, vault=vault)
                        if argument_type == "rpe_var":
                            try:
                                var = retrive_variable(argument, routine=current_block, vault=vault)
                            except AssertionError:
                                continue
                            insert_in_line(var, source_file, counter, _index)
                        if argument_type == "real" and argument.count("%val"):
                            try:
                                m = re.search("(\w*) *%val", argument)
                                var_name = m.groups(0)[0]
                                if not var_name.strip():
                                    var_name = argument
                            except AttributeError:
                                var_name = argument
                            var = retrive_variable(var_name, routine=current_block, vault=vault)
                            insert_in_line(var, source_file, counter, _index)


def insert_in_line(var, source_file, counter, line_index):
    sbits_assignation = "    %s%%sbits = emulator_variable_precisions(%i)"
    apply_truncation = "    CALL apply_truncation(%s) "
    if not var.id:
        var.id = counter.count
        counter.up()
    source_file.lines.insert(line_index + 1, sbits_assignation % (var.name, var.id))
    source_file.lines.insert(line_index + 2, apply_truncation % var.name)


def assign_sbits_to_rpe_variables(path_to_sources, vault=None):
    source_files = load_source_files(path_to_sources)
    rpe_variables = [var for var in vault.variables if var.type == "rpe_var" and not var.is_parameter]
    allocatable, non_allocatable = split_vars_by_allocatable(rpe_variables)
    var_counter = Counter()
    # Allocatables:
    for variable in allocatable:
        insert_variable_precision_specification_to_allocatable(variable, source_files, var_counter)

    # Non allocatables that appear in a namelist:
    na_in_namelist, na_not_in_namelist = split_by_appearance_in_namelist(non_allocatable, source_files)
    for variable in na_in_namelist:
        insert_variable_precision_specification_to_namelist_parameter(variable, source_files, var_counter)
    in_main, not_in_main = split_vars_by_main_or_routine(na_not_in_namelist)

    # Variables inside routines
    for variable in not_in_main:
        if variable.intent != "in":
            insert_variable_precision_to_routine_variable(variable, source_files, var_counter)

    non_identified_variables = [var for var in rpe_variables if var.id is None]
    print("Implementation ended!")
    print("From %i total variables %i do not appear in the namelist (usually routine inputs)." % (len(rpe_variables)),
          len(non_identified_variables))

    # To save the variable id into the vault.
    save = True
    if save:
        # Saving modified files
        for filename, source_file in source_files.items():
            source_file.save_modified_file(join(path_to_sources, filename))
        # This workaround is needed to allow pickle to write the file, don't know yet about its risks
        import sys
        sys.setrecursionlimit(2000)
        pickle.dump(vault, open("vault.pkl", "wb"))
    return var_counter


def create_read_precisions_module_and_add_to_file(path_to_sources, target_module=None, vault=None):
    # Create module
    rpe_vars = [var for var in vault.variables if var.type == "rpe_var"]
    assigned_rpe_vars = [var for var in rpe_vars if var.id is not None]
    # Path to template is hardcoded
    with open("AdditionalFiles/read_precisions_template", "r") as input_file:
        template = input_file.read()
    module = template.replace("%NUMBER_OF_RPE_VARS%", "%i" % len(assigned_rpe_vars))
    with open(join(path_to_sources, "read_precisions.f90"), "w") as output_file:
        output_file.write(module)
    # Put call into target module

    if target_module:
        if isfile(join(path_to_sources,target_module)):
            with open(join(path_to_sources, target_module), "r") as input_file:
                original_source = input_file.read()

            lines = original_source.split("\n")
            lines.insert(1, "use read_precisions")
            for index, line in enumerate(lines):
                if line.lower().strip() == "private":
                    lines.pop(index)
            original_source = "\n".join(lines)

            with open(join(path_to_sources, target_module), "w") as output_file:
                output_file.write(original_source)
        else:
            warnings.warn("File %s not found in sources.\nUnable to insert read_precisions module")
